# wwwd.rb
#
# TODO:
# * eruby integration
# * public_html simulation



require "mime.rb"
require "selectserver.rb"
require "log.rb"
include FileTest

def dir(path)
	files = []
	Dir.foreach(path){|item|
		next if item == '.' or item == '..'
		files << "#{path}/#{item}" if file?(item)
		files += dir("#{path}/#{item}") if directory?(item)
	}
	return files
end

def main
	ARGV.push( "wwwd.conf" ) if( ARGV.size == 0 ) 

	# Let's gather info.....
	port=80
	fileroot="/var/wwwd"
	bindaddr="0.0.0.0"
	maxcon=100
	client_timeout = 20
	maxreadbuf = 65536
	error_log_file = nil  
	access_log_file = nil  
	log_mode = "connection file script status internal"
	maxrequestlen = 8192   # HTTP max header length
	preload_modules = {}   # Modules configuration
	eval( File.open( ARGV.shift,"r").read )     # finally read config file

	# Initialize everything.
	log = Log.new( access_log_file, error_log_file, log_mode )
	hss = HTTPSelectServer.new( fileroot, preload_modules, maxrequestlen,
							   bindaddr, port, maxcon, client_timeout, 
							   maxreadbuf, log, false )
	# Let's fly.
	hss.forever 
end

class HTTPSelectServer < SelectServer
	include MimeTypes

	def initialize( fileroot, premods, maxreql,
				   bindaddr,port,maxcon,client_timeout, maxreadbuf,log,revf)
		super( bindaddr,port,maxcon,client_timeout,maxreadbuf,log,revf)
		@log = log
		@fileroot = fileroot
		@preload_modules = premods
		@maxrequestlen = maxreql
		
		@h_file_mem = {}       # save file into this.
		@h_script_obj = {}     # save compiled scripts into this 


		# Prepare files loaded
		count = 0
#		`find #{fileroot} -type f`.each{ |i|
		dir(fileroot).each{ |i|
			count += 1
			i.strip!
			suf = File.basename(i).split(/\./).pop
			if( MIME_HASH[suf] ) then
				@h_file_mem[ i.sub(fileroot,"") ] = File.open(i,"r").read
#				print i,"\n"
			end
		}
		@log.print( "server", self.to_s, "loaded #{count} files" )

		# Prepare all scripts loaded, and compiled
		count = 0
		@preload_modules.keys.each { |m|
			count+=1
			raise "No '/' on the top of location :#{m}\n" if !(m=~/^\// )
			script = File.open( @preload_modules[m],"r").read
			@h_script_obj[m] = Object.new
			eval( "class << @h_script_obj[m]\n#{script}\nend\n" )
		}
		@log.print( "server",self.to_s, "compiled #{count} scripts" )

		GC.start # To ensure files are closed.
	end
	def make_http_reply( code, name, mime, content )
		l = content.size
		a = "HTTP/1.0 #{code} #{name}\r\n"
		a+= "Server: wwwd.rb\r\n"
		a+= "Content-type: #{mime}\r\n"
		a+= "Content-length: #{l}\r\n"
		a+= "\r\n"
		a+= content
		return a
	end
	def make_http_premature_header( code, name )
		a = "HTTP/1.0 #{code} #{name}\r\n"
		a+= "Server: wwwd.rb\r\n"
		return a
	end
	def make_error_html( title, mesg )
		a = "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
		a+= "<HTML><HEAD>\n"
		a+= "<TITLE>#{title}</TITLE>\n"
		a+= "</HEAD><BODY>\n"
		a+= "<H1>#{title}</H1>\n"
		a+= "<B>#{mesg}</B>\n"
		a+= "</BODY></HTML>\n"
		return a
	end

	# This method is expected by selectserver to parse protocol.
	def readbuf_check( peer, rb )
		lines = rb.split(/\n/)
		return [nil,false] if( lines.size < 2 )
		topline = lines.shift.tr("\r","").tr("\n","")
		reqtk = topline.split(/ /)
		if( reqtk.size != 3 ) then
			html = make_error_html( "400 Bad Request" , topline )
			reply = make_http_reply( 400, "Bad Request", "text/html", html )
			@log.print("http",peer.to_s,"400 totally bad request" ) if @log
			return [reply,true]
		end
		# Check if rb has empty line.
		has_empty_line=0
		reqheaders = []
		lines.each { |l|
			l.tr!("\r","") if l != nil
			l.tr!("\n","") if l != nil
			has_empty_line = 1 if( l == "" )
			reqheaders.push( l )
		}
		if( has_empty_line == 0 ) then
			@log.print("http",peer.to_s,"request header too long") if @log
			raise "request header too long" if( rb.size > @maxrequestlen ) 
		end
		
		# Analyze top line
		method = reqtk[0]
		uri = reqtk[1]
		ver = reqtk[2]
		# Analyze headers
		contentlen = -1
		user_agent = ""
		reqheaders.each { |r|
			if( r =~ /^content-length/i ) then
				contentlen = r.split(/ /).pop.to_i
			elsif( r =~ /^user_agent/i ) then
				tk = r.split(/ /).shift
				user_agent = tk.join( " " )
				ENV[ "USER_AGENT"] = user_agent
			else
				# Unknown request header.
			end
		}
		path , arg = uri.split( /\?/ )
		arg = "" if arg == nil		
		ENV["REQUEST_METHOD"] = method
		ENV["HTTP_VERSION"] = ver
		ENV["QUERY_STRING"] = arg
		
		# Now GET or POST. Check POST_TEXT and CONTENT_LENGTH.
		if( method == "GET" ) then
			ENV["POST_TEXT"] = ""
			ENV["CONTENT_LENGTH"] = arg.size.to_s
		elsif( method == "POST" ) then
			if( contentlen == -1 ) then
				html = make_error_html( "403 Forbidden" , 
									   "no content-len when POST" )
				reply = make_http_reply( 403, "Forbidden", "text/html", html)
				@log.print( "http",peer.to_s,
						   "403 no content-len posting to #{path}") if @log
				return [reply,true]
			end
print rb
			startind1 = rb.index( "\r\n\r\n" ) 
			startind2 = rb.index( "\n\n" ) 
			startind1 += 4 if( startind1 ) 
			startind2 += 2 if( startind2 )

			# Use smaller one, or non-nil one. safer way.

			if ( startind1 == nil ) then
				startind = startind2
			elsif( startind2 == nil ) then
				startind = startind1
			elsif( startind1 > startind2 ) then
				startind = startind2
			else
				startind = startind1
			end
			final_length= startind + contentlen.to_i 
			if( rb.size < final_length ) then
				return [nil,false]
			end
			final_length -= 1
			ENV["CONTENT_LENGTH"] = contentlen.to_s
			ENV["POST_TEXT"] = rb[startind..final_length]
		else
			html = make_error_html( "501 Not Implemented" , method )
			reply = make_http_reply( 501, "Not Implemented" ,"text/html",html)
			@log.print("http",peer.to_s,"501 method #{method}" ) if @log
			return [reply,true]
		end
		
		# Now file or script.
		
		if( @h_file_mem[path] != nil ) then          # Access to files
			suf = path.split(/\./).pop
			h = make_http_premature_header( 200, "OK" )
			h += "Content-length: #{@h_file_mem[path].size}\r\n"
			mimetype = MIME_HASH[suf]
			mimetype = "text/plain" if( mimetype == nil ) 
			h += "Content-type: #{mimetype}\r\n"
			h += "\r\n"
			h += @h_file_mem[path]
			@log.print( "file",peer.to_s,"200 #{path}" ) if @log
			return [h,true]
		elsif( @h_script_obj[path] != nil ) then  # Access to scripts
			ENV["FILE_PATH"] = path
			begin   # to catch any exceptions in external scripts
				h = make_http_premature_header( 200, "OK" )
				h += @h_script_obj[path].process
				@log.print( "script",peer.to_s,"200 #{path}" ) if @log
				return [h,true]
			rescue
#				html = make_error_html( 500, "Internal Server Error")
				html = make_error_html( 500, "Internal Server Error\n<PRE>#{$!}\n#{$@}</PRE>")
				reply = make_http_reply( 500, "Internal Server Error",
									   "text/html" , html )
				@log.print( "script",peer.to_s,"500 script error" ) if @log
				return [reply,true]
			end
		else
			# Not Found
			html = make_error_html( "404 Not Found" , 
								   "Sorry, #{path} is not found "+ 
								   "on this server." )
			reply = make_http_reply( 404, "Not Found", "text/html", html )
			@log.print( "file",peer.to_s,"404 #{path}" ) if @log
			return [reply,true]
		end
	end
end



main()
