#
# SelectServer : server class, uses select() 

# TODO
# * connection timeout
# * Access control by ruby-script filter
# * faster, better! : ( Pay attention on ruby-object allocation )

require "socket"

class Peer
	READ_UNIT = 4096
	def initialize(sock,maxreadbuf,server,log)
		@sock = sock
		@server = server
		@log = log
		@maxreadbuf = maxreadbuf
		@close_when_buffer_clear = false
		@rb = ""
		@wb = ""
		@peerval = Hash.new  # for application's use ( stateful protocol)

		# Notify application by new_connection method
		# retarray[0]  : message to be sent
		# retarray[1]  : close_if_clear if true
		# retarray[2]  : not used
		begin
			if defined? @server.new_connection then
				retarray = @server.new_connection(self)
				if( retarray.type == Array and retarray.size >= 2 ) then
					@close_when_buffer_clear = 1 if( retarray[1] )
					@wb += retarray[0] if ( retarray[0] )
				else
					raise "new_connection has to return [str,flg]"
				end
			end
		rescue
			reset( "initialize" )
			raise "new_connection fail"
		end
		@log.print( "connection", self.to_s , "new" ) if @log
	end
	def setindex(i)
		@i = i
	end
	def getindex
		return @i
	end
	def used()
		return @used
	end
	def sock()
		return @sock
	end
	def reset(msg)
		begin
			@log.print( "connection", self, "reset: #{msg}" ) if @log 
			if defined? @server.reset_connection
				@server.reset_connection(self)
			end
		rescue
			@log.print( "internal",self,"reset: #{msg} #{$!}") if @log
		end
		@sock.shutdown 2
		@sock.close
	end
	def suck
		begin
			r = @sock.sysread( READ_UNIT )
			@rb += r
			if( @rb.size > @maxreadbuf ) then
				reset( "suck-1" )
				return false
			end

			begin
				# Yeah, this is the only place to check read buffer.
				# In SelectServer, the read buffer is checked only when
				# data is comming from network.
				# retarray[0] : message to be sent
				# retarray[1] : close_if_clear if true
				# retarray[2] : set readbuf if String ( default false)
				retarray = @server.readbuf_check( self, @rb )
				if( retarray.type == Array and retarray.size >= 2 ) then
					@close_when_buffer_clear = 1 if( retarray[1] ) 
					rc = retarray[0]
				else
					raise "readbuf_check has to return [str,flg,<RB>]\n"
				end
				if( retarray.size == 3 and retarray[2] and 
				   retarray[2].type == String ) then
					@rb = retarray[2]
				end
			rescue
				STDERR.print "suck fail (readbuf_chcek #{$!})"
				@log.print( "internal", self, 
						   "suck fail (readbuf_check #{$!})") if @log
				reset( "suck-2" )
				return false
			end
			
			if( rc == nil ) then
				return true
			else
				@wb += rc
				return true
			end
		rescue
			reset( "suck-3" )
			return false
		end
	end
	def to_be_flushed
		return ( @wb.size > 0 ) 
	end
	def flush
		r = @sock.syswrite @wb
		if( !r or r == 0 ) then
			reset( "flush-1" )
			return false
		else
			size = @wb.size
			@wb = @wb[r..size]
			@wb = "" unless @wb
			
			if( @close_when_buffer_clear and @wb.size == 0 ) then
				reset( "flush-2" )
				return false
			end
			return true
		end
	end
	def store ( key, val )
		@peerval[key] = val
	end
	def [](k)
		return @peerval[k]
	end
	def []=(k,value)
		@peerval[k]=value
	end
	def to_s
		begin
			a = @sock.getpeername.unpack("nnN")
			return sprintf( "0x%08x:%d", a[2] ,a[1] )
		rescue
			return "getpeername fail: #{$!}"
		end
	end
end

class SelectServer

	# ps  : reference to class ProtocolSpec
	# log : reference to class Log
	# revflag : do reverse-name-resolve if true
	def initialize( bindaddr, port, maxcon, timeo, maxreadbuf, log, revflag )
		
		# Test successor
		if !defined? readbuf_check then
			raise "SelectServer needs successors to have readbuf_check()."
		end
		@maxcon = maxcon
		@log = log
		@maxreadbuf = maxreadbuf 
		@currentcon = 0
		@timeout_sec = timeo
		@revflag = revflag
		@mainsock = TCPServer.open( bindaddr, port )

		@peers = []
		maxcon.times{ |i|
			@peers[i] = nil
		}
		trap( "SIGPIPE" ) {}
	end

	def forever
		@log.print( "server", self.to_s, "started" ) if @log
		@now_old = 0
		while TRUE
			now = Time.now.to_i
			if( now != @now_old ) then
				@now_old = now
				if defined? each_second then
					each_second()
				end
			end

			readtest = []
			writetest = []
			readable = []
			writable = []

			readtest.push( @mainsock )
			sockhash = Hash.new
			@maxcon.times{ |i|
				if( @peers[i] != nil ) then
					readtest.push( @peers[i].sock )
					if( @peers[i].to_be_flushed ) then
						writetest.push( @peers[i].sock )
					end
					sockhash[ @peers[i].sock ] = @peers[i]
				end
			}
			q = IO.select( readtest, writetest, nil , 1 )
			next if( !q ) 

			q[0].each{|hoge| readable.push hoge	}
			q[1].each{|hoge| writable.push hoge	}

			readable.each { |rsock|
				if( rsock == @mainsock ) then
					newsock = @mainsock.accept()
					if( @currentcon == @maxcon ) then
						newsock.close
						next
					end
					begin
						newpeer = Peer.new( newsock, @maxreadbuf, self,@log )
						@maxcon.times{|i|
							if( @peers[i] == nil ) then
								@peers[i] = newpeer
								newpeer.setindex(i)
								@currentcon += 1
								break
							end
						}
					rescue
						@log.print( "internal","",
								   "something wrong (newpeer): #{$!}" )
					end
				else
					if( sockhash[rsock].suck == false ) then
						@currentcon -= 1
						@peers[ sockhash[rsock].getindex ] = nil
						sockhash[rsock]=nil
					end
				end
			}

			writable.each { |wsock|
				if( sockhash[wsock] and sockhash[wsock].flush == false ) then
					@currentcon -= 1
					@peers[sockhash[wsock].getindex ] = nil
					sockhash[wsock]=nil
				end
			}
		end
	end
	def to_s
		begin
			a = @mainsock.getsockname.unpack("nnN")
			return sprintf( "0x%08x:%d", a[2], a[1] )
		rescue
			return "getsockname fail: #{$!}"
		end
	end

end
